/*
 * Driver for Comtrol RocketPort EXPRESS/INFINITY cards
 *
 * Copyright (C) 2013 Comtrol Corp.
 *
 * Based on rp2.c by Kevin Cernekee <cernekee@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation.
 */

#include <linux/bitops.h>
#include <linux/compiler.h>
#include <linux/completion.h>
#include <linux/console.h>
#include <linux/delay.h>
#include <linux/firmware.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/ioport.h>
#include <linux/irq.h>
#include <linux/kernel.h>
#include <linux/log2.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/serial.h>
#include <linux/serial_core.h>
#include <linux/slab.h>
#include <linux/sysrq.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/types.h>

#define TraceLine(fmt, ...) printk(KERN_INFO "rocket: Trace %s():%d " fmt "\n", __FUNCTION__, __LINE__, ##__VA_ARGS__)

#define TX_FIFO_SIZE  255
#define RX_FIFO_SIZE 1023

#define RP_PCI_9030_GPIO_CTRL 0x54  // offset from bar1 for quad/octa detect input
#define RP_PCI_9030_RING_IND  0xc0  // offset from bar1 for ring-indicator inputs

//======================================================================
// start of backwards compatibilty stuff
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25)
# error "======================================================="
# error "== Sorry: kernel too old -- 2.6.25 or newer required =="
# error "======================================================="
#endif

#if !defined(BIT)
# define BIT(n) (1<<(n))
#endif

#if !defined PORT_RP
# define PORT_RP 103
#endif

#if !defined(CONFIG_SERIAL_RP_NR_UARTS)
# define CONFIG_SERIAL_RP_NR_UARTS 128
#endif

#if !defined(for_each_set_bit)
# define for_each_set_bit(bit, addr, size)                 \
    for ((bit) = find_first_bit((addr), (size));            \
       (bit) < (size);                                    \
       (bit) = find_next_bit((addr), (size), (bit) + 1))
#endif

#if !defined DEFINE_PCI_DEVICE_TABLE
#define DEFINE_PCI_DEVICE_TABLE(_table) \
    const struct pci_device_id _table[] __devinitdata
#endif

// macros to handle things that have moved around in the serial_core/tty API
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,32)
# define rp_state(rp)           rp->port.info
# define rp_tty(rp)             rp->port.info->tty
# define rp_delta_msr_wait(rp)  rp->port.info->delta_msr_wait
#else
# define rp_state(rp)           rp->port.state
# define rp_tty(rp)             rp->port.state->port.tty
# define rp_delta_msr_wait(rp)  rp->port.state->port.delta_msr_wait
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,9,0)
# define rp_tty_prepare_flip_string_flags(rp,cbuf,fbuf,cnt) tty_prepare_flip_string_flags(rp_tty(rp),cbuf,fbuf,cnt)
# define rp_tty_flip_buffer_push(rp) tty_flip_buffer_push(rp_tty(rp))
# define rp_tty_buffer_request_room(tty,size)        tty_buffer_request_room(tty,size)
# define rp_tty_insert_flip_string(tty,cbuf,size)    tty_insert_flip_string(tty,cbuf,size)
# define rp_tty_insert_flip_string_flags(tty,cbuf,fbuf,size) tty_insert_flip_string_flags(tty,cbuf,fbuf,size)
#else
# define rp_tty_buffer_request_room(tty,size)        tty_buffer_request_room(tty->port,size)
# define rp_tty_insert_flip_string(tty,cbuf,size)    tty_insert_flip_string(tty->port,cbuf,size)
# define rp_tty_insert_flip_string_flags(tty,cbuf,fbuf,size) tty_insert_flip_string_flags(tty->port,cbuf,fbuf,size)
# define rp_tty_prepare_flip_string_flags(rp,cbuf,fbuf,cnt) tty_prepare_flip_string_flags(&rp->port.state->port,cbuf,fbuf,cnt)
# define rp_tty_flip_buffer_push(rp) tty_flip_buffer_push(&rp->port.state->port)
#endif

// end of backwards compatibility stuff
//======================================================================

//======================================================================
// When adding a new board, it must be added to _both_ tables below:
// the model info table containing the board description and other
// info, and the PCI device table used to register the driver with the
// kernel.  For UPCI boards, the PCI device ID is not always identical
// to the model ID.
//
// For old PCI boards (model < 0x0800) model ID is same as PCI device
// ID.  For _some_ uPCI boards that can accept either quad or octa
// cables (0x0805, 0x80b, 0x812), if a quad cable is detected, we
// morph the model ID by setting bit 15 (0x08005 => 0x8805).

static const struct rp_model_info
{
  unsigned model_id;
  const char *description;
  int num_aiops;
  int num_ports;
  bool is_upci;
  int modem;      // 0==no, 1==RM-I, 2==RM-II, 3==RM-III, ...
  bool is_plus;
  bool quadocta;
  bool has_RI;
} rp_model_table[] =
{
  //                                                                    sw detect
  //model    description                   aiops ports upci modem plus  quad/octa   RI
  {0x0001,  "RP PCI 32 port external I/F",    4,  32,    0,   0,    0,      0,      0},
  {0x0002,  "RP PCI 8 port external I/F",     1,   8,    0,   0,    0,      0,      0},
  {0x0003,  "RP PCI 16 port external I/F",    2,  16,    0,   0,    0,      0,      0},
  {0x0004,  "RP PCI 4 port quad cable",       1,   4,    0,   0,    0,      0,      0},
  {0x0005,  "RP PCI 8 port octa cable",       1,   8,    0,   0,    0,      0,      0},
  {0x0006,  "RP PCI 8 port RJ11",             1,   8,    0,   0,    0,      0,      0},
  {0x0007,  "RP PCI 4 port RJ45",             1,   4,    0,   0,    0,      0,      0},
  {0x0008,  "RP PCI 8 port DB78 SNI",         1,   8,    0,   0,    0,      0,      0},
  {0x0009,  "RP PCI 16 port DB78 SNI",        2,  16,    0,   0,    0,      0,      0},
  {0x000a,  "RP PCI Plus 4 port",             1,   4,    0,   0,    1,      0,      1},
  {0x000b,  "RP PCI Plus 8 port",             2,   8,    0,   0,    1,      0,      1},
  {0x000c,  "RP PCI RocketModem II 6 port",   1,   6,    0,   2,    0,      0,      0},
  {0x000d,  "RP PCI RocketModem II 4 port",   1,   4,    0,   2,    0,      0,      0},
  {0x400c,  "RP PCI RocketModem IIs 6 port",  1,   6,    0,   2,    0,      0,      0},
  {0x400d,  "RP PCI RocketModem IIs 4 port",  1,   4,    0,   2,    0,      0,      0},
  {0x000e,  "RP PCI Plus 2 port RS232",       1,   2,    0,   0,    1,      0,      1},
  {0x000f,  "RP PCI Plus 2 port RS422",       1,   2,    0,   0,    1,      0,      1},

  {0x0801,  "RP UPCI 32 port external I/F",   4,  32,    1,   0,    0,      0,      0},
  {0x0802,  "RP UPCI 8 port external I/F",    1,   8,    1,   0,    0,      0,      0},
  {0x0803,  "RP UPCI 16 port external I/F",   2,  16,    1,   0,    0,      0,      0},
  {0x0805,  "RP UPCI 8 port octa cable",      1,   8,    1,   0,    0,      1,      1},
  {0x8805,  "RP UPCI 4 port quad cable",      1,   4,    1,   0,    0,      1,      1},
  {0x080b,  "RP UPCI Plus 8 port",            2,   8,    1,   0,    1,      1,      1},
  {0x880b,  "RP UPCI Plus 4 port",            1,   4,    1,   0,    1,      1,      1},
  {0x080c,  "RP UPCI Rocketmodem III 8 port", 1,   8,    1,   3,    0,      0,      1},
  {0x080d,  "RP UPCI Rocketmodem III 4 port", 1,   4,    1,   3,    0,      0,      1},
  {0x080e,  "RP UPCI Plus 2 port RS232",      1,   2,    1,   0,    1,      0,      1},
  {0x080f,  "RP UPCI Plus 2 port RS422",      1,   2,    1,   0,    1,      0,      1},
  {0x0810,  "RP UPCI Plus 4 port RJ45",       1,   4,    1,   0,    1,      0,      1},
  {0x0811,  "RP UPCI Plus 8 port RJ11",       2,   8,    1,   0,    1,      0,      1},
  {0x0812,  "RP UPCI Plus 8 port RS422",      2,   8,    1,   0,    1,      1,      1},
  {0x8812,  "RP UPCI Plus 4 port RS422",      1,   4,    1,   0,    1,      1,      1},
  {0x0813,  "RP UPCI Rocketmodem IV 8 port",  1,   8,    1,   4,    0,      0,      1},
  {0x0814,  "RP UPCI Rocketmodem IV 4 port",  1,   4,    1,   4,    0,      0,      1},
  {0x0903,  "RP CPCI 16 port external I/F",   2,  16,    1,   0,    0,      0,      0},
  {}
};

static const struct rp_model_info *get_model_info(unsigned model_id)
{
  const struct rp_model_info *p;
  for (p=rp_model_table; p->model_id; ++p)
    if (model_id == p->model_id)
      return p;
  return NULL;
}

#define RP_ID(prod) PCI_VDEVICE(RP, (prod))

static DEFINE_PCI_DEVICE_TABLE(rp_pci_tbl) =
{
  // PCI ID

  // old PCI boards
  {RP_ID(0x0001)},   //RP 32 port w/external I/F     
  {RP_ID(0x0002)},   //RP 8 port w/external I/F      
  {RP_ID(0x0003)},   //RP 16 port w/external I/F     
  {RP_ID(0x0004)},   //RP 4 port w/quad cable        
  {RP_ID(0x0005)},   //RP 8 port w/octa cable        
  {RP_ID(0x0006)},   //RP 8 port w/RJ11 connectors   
  {RP_ID(0x0007)},   //RP 4 port w/RJ11 connectors   
  {RP_ID(0x0008)},   //RP 8 port w/ DB78 SNI (Siemens) connector 
  {RP_ID(0x0009)},   //RP 16 port w/ DB78 SNI (Siemens) connector   
  {RP_ID(0x000a)},   //RP Plus 4 port                
  {RP_ID(0x000b)},   //RP Plus 8 port                
  {RP_ID(0x000c)},   //RM II 6 port                    
  {RP_ID(0x000d)},   //RM II 4 port                    
  {RP_ID(0x000e)},   //RP Plus 2 port RS232          
  {RP_ID(0x000f)},   //RP Plus 2 port RS422           

  // new UPCI boards
  {RP_ID(0x0801)},   //RP UPCI 32 port w/external I/F  
  {RP_ID(0x0802)},   //RP UPCI 8 port w/external I/F  
  {RP_ID(0x0803)},   //RP UPCI 16 port w/external I/F 
  {RP_ID(0x0805)},   //RP UPCI 8 port quad/octa
  {RP_ID(0x080b)},   //RP UPCI Plus quad/octa
  {RP_ID(0x080c)},   //RM III 8 port                 
  {RP_ID(0x080d)},   //RM III 4 port                 
  {RP_ID(0x080e)},   //RP UPCI Plus 2 port RS232
  {RP_ID(0x080f)},   //RP UPCI Plus 2 port RS422
  {RP_ID(0x0810)},   //RP UPCI Plus 4 port w/RJ45 connectors
  {RP_ID(0x0811)},   //RP UPCI Plus 8 port w/RJ11 connectors
  {RP_ID(0x0812)},   //RP UPCI Plus 8 port RS422
  {RP_ID(0x0813)},   //RP UPCI RM IV 8 port
  {RP_ID(0x0814)},   //RP UPCI RM IV 4 port
  // Compact PCI boards
  {RP_ID(0x0903)},   //RP Compact PCI 16 port w/external I/F 
  {}
};

MODULE_DEVICE_TABLE(pci, rp_pci_tbl);


//======================================================================
// To avoid problems with old distros/kernels, use internal firmware blob

static u8 rp_firmware_blob[] =
{
  /*0x900*/   0xf6, 0x82, 0x86, 0xfb, 0x00, 0x0a, 0x01, 0x0a, 
  /*0x908*/   0x8a, 0x13, 0xc5, 0x11, 0x86, 0x85, 0x20, 0x0a,
  /*0x910*/   0x21, 0x0a, 0x41, 0xff, 0x82, 0x00, 0x82, 0x7b,
  /*0x918*/   0x8a, 0x7d, 0x88, 0x81, 0x86, 0x7a, 0x84, 0x81,
  /*0x920*/   0x82, 0x7c, 0x08, 0x0a
};
static struct firmware rp_firmware =
{
  .size = sizeof rp_firmware_blob,
  .data = rp_firmware_blob,
};
//
//======================================================================

#define DRV_VERS             "4.01"
#define DRV_NAME              "rocket"

// RocketPort Plus boards have a higher clock frequency so they can
// handle faster baud rates -- therefore plus boards require different
// prescalers and divisors

#define RP_PRESCALE           0x14
#define RP_BAUD_BASE          460800
#define RP_DEF_BAUD_DIV       (RP_BAUD_BASE/9600)

#define RP_PLUS_PRESCALE      0x12
#define RP_PLUS_BAUD_BASE     921600
#define RP_PLUS_DEF_BAUD_DIV  (RP_PLUS_BAUD_BASE/9600)

// Internal character status bit that we use to let uart_insert_char()
// drop bytes received on a !CREAD port
#define RP_DUMMY_READ                   BIT(16)


//======================================================================
//  AIOPIC register offsets
//
//Direct access registers
//
//  Global 
#define RP_CMD_REG   0x38   // Command Register            8    Write
#define RP_INT_CHAN  0x39   // Interrupt Channel Register  8    Read
#define RP_INT_MASK  0x3A   // Interrupt Mask Register     8    Read / Write
#define RP_INDX_ADDR 0x3C   // Index Register Address      16
#define RP_INDX_DATA 0x3E   // Index Register Data         8/16 Read / Write

//  Channel
#define RP_CHN_DATA  0x00   // Tx/Rx Data                  16   Write
#define RP_CHN_STAT  0x20   // Channel Status              8/16 Read / Write
#define RP_FIFO_CNT  0x10   // Transmit/Receive FIFO Count 16   Read
#define RP_INT_ID    0x30   // Interrupt Identification    8    Read

//Indirect access registers (one copy per channel, spacing==0x1000
#define RP_TX_CTRL           0xFF0  // Transmit Control               16  Write
#define RP_RX_CTRL           0xFF2  // Receive Control                 8  Write
#define RP_BAUD              0xFF4  // Baud Rate                      16  Write
#define RP_CLK_PRE           0xFF6  // Clock Prescaler                 8  Write

#define RP_TX_ENBLS          0x980  // Tx Processor Enables Register 8 Read / Write
#define RP_TXCMP1            0x988  // Transmit Compare Value #1     8 Read / Write
#define RP_TXCMP2            0x989  // Transmit Compare Value #2     8 Read / Write
#define RP_TXREP1B1          0x98A  // Tx Replace Value #1 - Byte 1  8 Read / Write
#define RP_TXREP1B2          0x98B  // Tx Replace Value #1 - Byte 2  8 Read / Write */
#define RP_TXREP2            0x98C  // Transmit Replace Value #2     8 Read / Write */


#define RP_RXFIFO_OUT_PTR    0x990  // rx fifo output pointer (16-bit)
#define RP_RXFIFO_IN_PTR     0x992  // rx fifo input pointer (16-bit)
#define RP_TXFIFO_OUT_PTR    0x994  // tx fifo output pointer (8-bit)
#define RP_TXFIFO_IN_PTR     0x995  // tx fifo input pointer (8-bit)
#define RP_TXPRI_CNT         0x996  // tx priority count
#define RP_TXPRI_PTR         0x997  // tx priority pointer

//  bytes in microcode that are modified to do stuff
#define RP_RXPROC_STARTSTOP  0x900  // f6==start,  0a==stop
#define RP_RXFIFO_EN         0x922  // 08==enable, 0a==disable


//======================================================================
// Bit definitions for registers

// Channel Data upper status bits (when data reg is read in status mode)
#define RP_CHN_DATA_STATMODE  0x8000    /* status mode enable bit */
#define RP_CHN_DATA_RXFOVERFL 0x2000    /* receive FIFO overflow */
#define RP_CHN_DATA_RX2MATCH  0x1000    /* receive compare byte 2 match */
#define RP_CHN_DATA_RX1MATCH  0x0800    /* receive compare byte 1 match */
#define RP_CHN_DATA_RXBREAK   0x0400    /* received BREAK */
#define RP_CHN_DATA_RXFRAME   0x0200    /* received framing error */
#define RP_CHN_DATA_RXPARITY  0x0100    /* received parity error */
#define RP_CHN_DATA_STATERROR (RP_CHN_DATA_RXBREAK | RP_CHN_DATA_RXFRAME | RP_CHN_DATA_RXPARITY | RP_CHN_DATA_RXFOVERFL)

// Interupt ID register values
#define RP_INTID_RXF_TRIG  0x20         /* Rx FIFO trigger level interrupt */
#define RP_INTID_TXFIFO_MT 0x10         /* Tx FIFO empty interrupt */
#define RP_INTID_SRC_INT   0x08         /* special receive condition interrupt */
#define RP_INTID_DELTA_CD  0x04         /* CD change interrupt */
#define RP_INTID_DELTA_CTS 0x02         /* CTS change interrupt */
#define RP_INTID_DELTA_DSR 0x01         /* DSR change interrupt */

// Channel Status bit definitions
#define RP_CHN_STAT_RDA       0x0001 // rx data available
#define RP_CHN_STAT_TXSHRMT   0x0002 // tx shift register empty
#define RP_CHN_STAT_TXFIFOMT  0x0004 // tx FIFO empty
#define RP_CHN_STAT_CD        0x0008 // CD state
#define RP_CHN_STAT_DSR       0x0010 // DSR state
#define RP_CHN_STAT_CTS       0x0020 // CTS state
#define RP_CHN_STAT_RXPARITY  0x0100 // rx parity error
#define RP_CHN_STAT_RXFRAME   0x0200 // rx framing error
#define RP_CHN_STAT_RXBREAK   0x0400 // rx BREAK
#define RP_CHN_STAT_RX1MATCH  0x0800 // rx compare byte 1 match
#define RP_CHN_STAT_RX2MATCH  0x1000 // rx compare byte 2 match
#define RP_CHN_STAT_RXFOVERFL 0x2000 // rx FIFO overflow
#define RP_CHN_STAT_STATMODE  0x8000 // status mode enable bit
#define RP_CHN_STAT_RXERROR   (RP_CHN_STAT_RXBREAK | RP_CHN_STAT_RXFRAME | RP_CHN_STAT_RXPARITY)
#define RP_CHN_STAT_TXDRAINED (RP_CHN_STAT_TXFIFOMT | RP_CHN_STAT_TXSHRMT)

// Command register bit definitions
#define RP_CMD_RESET_ALL 0x80   // reset AIOP (all channels)
#define RP_CMD_TXOVERIDE 0x40   // Transmit software off override
#define RP_CMD_RESETUART 0x20   // reset channel's UART
#define RP_CMD_RESTXFCNT 0x10   // reset channel's Tx FIFO count register
#define RP_CMD_RESRXFCNT 0x08   // reset channel's Rx FIFO count register

// Tx Control bit definitions
#define RP_TX_CTRL_DATA8BIT  0x0001  // 8 bit data (0 = 7 bit data)
#define RP_TX_CTRL_EVEN_PAR  0x0002  // even parity (0 = odd parity)
#define RP_TX_CTRL_PARITY_EN 0x0004  // enable parity (0 = no parity)
#define RP_TX_CTRL_STOP2     0x0008  // enable 2 stop bits (0 = 1 stop)
#define RP_TX_CTRL_TXINT_EN  0x0010  // transmit interrupt enable
#define RP_TX_CTRL_RTSTOG_EN 0x0040  // RTS toggle enable bit
#define RP_TX_CTRL_CTSFC_EN  0x0080  // CTS flow control enable bit
#define RP_TX_CTRL_TX_ENABLE 0x0100  // enable transmitter
#define RP_TX_CTRL_SET_RTS   0x0200  // assert RTS
#define RP_TX_CTRL_SET_DTR   0x0400  // assert DTR
#define RP_TX_CTRL_LOCALLOOP 0x0800  // local loopback set for test
#define RP_TX_CTRL_SETBREAK  0x1000  // send break condition (must clear)

// Rx Control bit definitions
#define RP_RX_CTRL_MCINT_EN  0x01  // modem change interrupt enable
#define RP_RX_CTRL_RXINT_EN  0x02  // Rx interrupt enable
#define RP_RX_CTRL_SRCINT_EN 0x04  // special Rx condition interrupt enable
#define RP_RX_CTRL_TRIG_1    0x08  // trigger level 1 char
#define RP_RX_CTRL_TRIG_MASK 0x18  // trigger level mask
#define RP_RX_CTRL_TRIG_NO   0x00  // Rx FIFO trigger level 0 (no trigger)
#define RP_RX_CTRL_TRIG_1_2  0x10  // trigger level 1/2
#define RP_RX_CTRL_TRIG_7_8  0x18  // trigger level 7/8
#define RP_RX_CTRL_RXPROC_EN 0x20  // receive processor enable
#define RP_RX_CTRL_RTSFC_EN  0x40  // RTS flow control enable


static struct uart_driver rp_uart_driver =
{
  .owner        = THIS_MODULE,
  .driver_name  = DRV_NAME,
  .dev_name     = "ttyR",
  .nr           = CONFIG_SERIAL_RP_NR_UARTS,
};

struct rp_card;

struct rp_uart_port
{
  struct uart_port              port;
  int                           idx;   // port index within aiopic
  int                           active;
  const struct port_mode       *mode;
  struct rp_card               *card;
  spinlock_t                   *aiop_lock;
  unsigned long                 aiop_base;
  unsigned long                 data_reg;
  unsigned long                 status_reg;
  unsigned long                 fifocnt_reg;
  unsigned long                 intid_reg;
  unsigned long                 index_base;
  uint16_t                      tx_ctrl_value;
  uint8_t                       rx_ctrl_value;
  unsigned int                  mctrl;
  // temp buffers for receive data
  uint8_t                       cbuf[1024];
  uint8_t                       fbuf[1024];
};

static inline struct rp_uart_port *port_to_rp(struct uart_port *port)
{
  return container_of(port, struct rp_uart_port, port);
}

static inline unsigned rx_fifo_count(struct rp_uart_port *rp)
{
  return inw_p(rp->fifocnt_reg);
}

static inline unsigned tx_fifo_count(struct rp_uart_port *rp)
{
  return inb_p(rp->fifocnt_reg);
}


// Utility functions to read/write indirect-access registers.  There
// is a spinlock for each AIOPIC that we use to prevent race
// conditions when accessing indirect registers via the
// INDX_ADDR/INDX_DATA registers

static void inline rp_outw_indexed(struct rp_uart_port *rp, uint16_t reg, uint16_t val)
{
  unsigned long flags;
  spin_lock_irqsave(rp->aiop_lock, flags);
  outw_p(rp->index_base+reg, rp->aiop_base+RP_INDX_ADDR);
  outw_p(val,                rp->aiop_base+RP_INDX_DATA);
  spin_unlock_irqrestore(rp->aiop_lock,flags);
}

static void inline rp_outb_indexed(struct rp_uart_port *rp, uint16_t reg, uint8_t val)
{
  unsigned long flags;
  spin_lock_irqsave(rp->aiop_lock, flags);
  outw_p(rp->index_base+reg, rp->aiop_base+RP_INDX_ADDR);
  outb_p(val,                rp->aiop_base+RP_INDX_DATA);
  spin_unlock_irqrestore(rp->aiop_lock,flags);
}

static uint16_t inline rp_inw_indexed(struct rp_uart_port *rp, uint16_t reg, uint16_t val)
{
  uint16_t r;
  unsigned long flags;
  spin_lock_irqsave(rp->aiop_lock, flags);
  outw_p( rp->index_base+reg,  rp->aiop_base+RP_INDX_ADDR);
  r = inw_p(rp->aiop_base+RP_INDX_DATA);
  spin_unlock_irqrestore(rp->aiop_lock,flags);
  return r;
}

static uint8_t inline rp_inb_indexed(struct rp_uart_port *rp, uint16_t reg)
{
  uint8_t r;
  unsigned long flags;
  spin_lock_irqsave(rp->aiop_lock, flags);
  outw_p(rp->index_base+reg, rp->aiop_base+RP_INDX_ADDR);
  r = inb_p(rp->aiop_base+RP_INDX_DATA);
  spin_unlock_irqrestore(rp->aiop_lock,flags);
  return r;
}

// Utility functions used to set/clear bits in the tx and rx control
// registers.  These registers are write-only, so we have to keep a
// backup copy of the values we write.

static void set_tx_ctrl_bits(struct uart_port *port, uint16_t mask)
{
  struct rp_uart_port *rp = port_to_rp(port);
  if ((rp->tx_ctrl_value & mask) == mask)
    return;
  rp->tx_ctrl_value |= mask;
  rp_outw_indexed(rp, RP_TX_CTRL, rp->tx_ctrl_value);
}

static void clr_tx_ctrl_bits(struct uart_port *port, uint16_t mask)
{
  struct rp_uart_port *rp = port_to_rp(port);
  if ((rp->tx_ctrl_value & mask) == 0)
    return;
  rp->tx_ctrl_value &= ~mask;
  rp_outw_indexed(rp, RP_TX_CTRL, rp->tx_ctrl_value);
}

static void set_rx_ctrl_bits(struct uart_port *port, uint16_t mask)
{
  struct rp_uart_port *rp = port_to_rp(port);
  if ((rp->rx_ctrl_value & mask) == mask)
    return;
  rp->rx_ctrl_value |= mask;
  rp_outb_indexed(rp, RP_RX_CTRL, rp->rx_ctrl_value);
}

static void clr_rx_ctrl_bits(struct uart_port *port, uint16_t mask)
{
  struct rp_uart_port *rp = port_to_rp(port);
  if ((rp->rx_ctrl_value & mask) == 0)
    return;
  rp->rx_ctrl_value &= ~mask;
  rp_outb_indexed(rp, RP_RX_CTRL, rp->rx_ctrl_value);
}


struct rp_card
{
  struct pci_dev                *pdev;
  struct rp_uart_port           *ports;
  int                           n_ports;
  int                           initialized_ports;
  int                           minor_start;
  unsigned long                 aiop_base;
  unsigned long                 pci_config_base;
  spinlock_t                    aiop_lock[4];
  bool                          shutting_down;
  const struct rp_model_info   *model_info;
  struct completion             fw_loaded;
  struct timer_list             poll_timer;
};

#define RP_ID(prod) PCI_VDEVICE(RP, (prod))

static DEFINE_SPINLOCK(rp_minor_lock);
static int rp_minor_next;

static int polling_mode = 100;  // module command line parameter (polling frequency, in Hz)
static int polling_jiffies;     // jiffies per polling cycle

static int rp_alloc_ports(int n_ports)
{
  int ret = -ENOSPC;
  spin_lock(&rp_minor_lock);
  if (rp_minor_next + n_ports <= CONFIG_SERIAL_RP_NR_UARTS)
    {
      ret = rp_minor_next;
      rp_minor_next += n_ports;
    }
  spin_unlock(&rp_minor_lock);
  return ret;
}

static void card_modem_reset(struct rp_card *card)
{
  unsigned i;
  for (i=0; i<card->n_ports; ++i)
    outb_p(0,card->aiop_base+0x40+i);
  msleep(500);
  for (i=0; i<card->n_ports; ++i)
    outb_p(0,card->aiop_base+0x48+i);
  msleep(500);
}

static void port_modem_reset(struct rp_uart_port *rp)
{
  outb_p(0,rp->aiop_base+0x40+rp->idx);
  msleep(500);
  outb_p(0,rp->aiop_base+0x48+rp->idx);
  msleep(500);
}

//======================================================================
// rp_uart_*****() functions are the methods provided to the
// serial_core layer.


static unsigned int rp_uart_tx_empty(struct uart_port *port)
{
  uint16_t status = inw_p(port_to_rp(port)->status_reg);
  if ((status & RP_CHN_STAT_TXDRAINED) == RP_CHN_STAT_TXDRAINED)
    return TIOCSER_TEMT;
  else
    return 0;
}

static inline unsigned get_ring_indicator(struct rp_uart_port *rp)
{
  if (rp->card->model_info->is_plus)
    return inw_p(rp->status_reg+8) & RP_CHN_STAT_DSR;
  else
    return !(inb_p(rp->card->aiop_base + RP_PCI_9030_RING_IND) & BIT(rp->idx));
}

static unsigned int rp_uart_get_mctrl(struct uart_port *port)
{
  struct rp_uart_port *rp = port_to_rp(port);
  unsigned ret = 0;
  uint16_t status = inw_p(rp->status_reg);
  if (status & RP_CHN_STAT_CD)  ret |= TIOCM_CAR;
  if (status & RP_CHN_STAT_DSR) ret |= TIOCM_DSR;
  if (status & RP_CHN_STAT_CTS) ret |= TIOCM_CTS;
  if (rp->card->model_info->has_RI && get_ring_indicator(rp)) ret |= TIOCM_RI;
  return ret;
}

static void rp_uart_set_mctrl(struct uart_port *port, unsigned int mctrl)
{
  struct rp_uart_port *rp = port_to_rp(port);
  rp->mctrl = mctrl;

  if (!(rp->tx_ctrl_value & RP_TX_CTRL_RTSTOG_EN) && !(rp->rx_ctrl_value & RP_RX_CTRL_RTSFC_EN))
    {
      if (mctrl & TIOCM_RTS)
        rp->tx_ctrl_value |= RP_TX_CTRL_SET_RTS;
      else
        rp->tx_ctrl_value &= ~RP_TX_CTRL_SET_RTS;
    }

  if (mctrl & TIOCM_DTR)
    rp->tx_ctrl_value |= RP_TX_CTRL_SET_DTR;
  else
    rp->tx_ctrl_value &= ~RP_TX_CTRL_SET_DTR;

  if (mctrl & TIOCM_LOOP)
    rp->tx_ctrl_value |= RP_TX_CTRL_LOCALLOOP;
  else
    rp->tx_ctrl_value &= ~RP_TX_CTRL_LOCALLOOP;

  rp_outw_indexed(rp, RP_TX_CTRL, rp->tx_ctrl_value);
}

static void rp_uart_start_tx(struct uart_port *port)
{
  set_tx_ctrl_bits(port,RP_TX_CTRL_TX_ENABLE);
}

static void rp_uart_stop_tx(struct uart_port *port)
{
  clr_tx_ctrl_bits(port,RP_TX_CTRL_TX_ENABLE);
}

static void rp_uart_stop_rx(struct uart_port *port)
{
  clr_rx_ctrl_bits(port,RP_RX_CTRL_RXPROC_EN);
}

static void rp_uart_break_ctl(struct uart_port *port, int break_state)
{
  if (break_state)
    set_tx_ctrl_bits(port,RP_TX_CTRL_SETBREAK);
  else
    clr_tx_ctrl_bits(port,RP_TX_CTRL_SETBREAK);
}

static void rp_uart_enable_ms(struct uart_port *port)
{
  set_rx_ctrl_bits(port,RP_RX_CTRL_MCINT_EN);
}

static void __rp_uart_set_termios(struct rp_uart_port *rp,
                                   unsigned long cfl,
                                   unsigned long ifl,
                                   cc_t *c_cc,
                                   unsigned int baud_div)
{
  // == called with port lock held ==

  // baud rate divisor (calculated elsewhere).  In hardware, 0 =
  // divide-by-1, so use divisor-1
  rp_outw_indexed(rp,RP_BAUD,baud_div-1);

  if (cfl & CSTOPB)
    rp->tx_ctrl_value |= RP_TX_CTRL_STOP2;
  else
    rp->tx_ctrl_value &= ~RP_TX_CTRL_STOP2;

  if ((cfl & CSIZE) == CS8)
    rp->tx_ctrl_value |= RP_TX_CTRL_DATA8BIT;
  else
    rp->tx_ctrl_value &= ~RP_TX_CTRL_DATA8BIT;

  if (cfl & PARENB)
    rp->tx_ctrl_value |= RP_TX_CTRL_PARITY_EN;
  else
    rp->tx_ctrl_value &= ~RP_TX_CTRL_PARITY_EN;

  if (cfl & PARODD)
    rp->tx_ctrl_value &= ~RP_TX_CTRL_EVEN_PAR;
  else
    rp->tx_ctrl_value |= RP_TX_CTRL_EVEN_PAR;

  if (cfl & CRTSCTS)
    {
      rp->rx_ctrl_value |= RP_RX_CTRL_RTSFC_EN;
      rp->tx_ctrl_value &= ~RP_TX_CTRL_SET_RTS;
      rp->tx_ctrl_value |= RP_TX_CTRL_CTSFC_EN;
    }
  else
    {
      rp->rx_ctrl_value &= ~RP_RX_CTRL_RTSFC_EN;
      rp->tx_ctrl_value &= ~RP_TX_CTRL_CTSFC_EN;
      if (rp->mctrl & TIOCM_RTS)
        rp->tx_ctrl_value |= RP_TX_CTRL_SET_RTS;
      else
        rp->tx_ctrl_value &= ~RP_TX_CTRL_SET_RTS;
    }

  rp_outw_indexed(rp, RP_TX_CTRL, rp->tx_ctrl_value);
  rp_outb_indexed(rp, RP_RX_CTRL, rp->rx_ctrl_value);

}

static void rp_uart_set_termios(struct uart_port *port,
                                struct ktermios *new,
                                struct ktermios *old)
{
  struct rp_uart_port *rp = port_to_rp(port);
  unsigned long flags;
  unsigned int baud, baud_div;

  baud = uart_get_baud_rate(port, new, old, 0, port->uartclk / 16);
  baud_div = uart_get_divisor(port, baud);

  if (tty_termios_baud_rate(new))
    tty_termios_encode_baud_rate(new, baud, baud);

  spin_lock_irqsave(&port->lock, flags);

  /* ignore all characters if CREAD is not set */
  port->ignore_status_mask = (new->c_cflag & CREAD) ? 0 : RP_DUMMY_READ;

  __rp_uart_set_termios(rp, new->c_cflag, new->c_iflag, new->c_cc, baud_div);
  uart_update_timeout(port, new->c_cflag, baud);

  spin_unlock_irqrestore(&port->lock, flags);
}


// Function called by polling timer callback to handle receive data

static void rp_rx_chars(struct rp_uart_port *rp)
{
  struct tty_struct *tty = rp_tty(rp);
  unsigned rxfifo, bytes, status;

  rxfifo = rx_fifo_count(rp);
  status = inw_p(rp->status_reg);
  bytes  = rp_tty_buffer_request_room(tty, rxfifo);

  if (bytes == 0)
    return;

  // need to enable status mode?
  if (status & (RP_CHN_STAT_RXPARITY | RP_CHN_STAT_RXFRAME | RP_CHN_STAT_RXBREAK | RP_CHN_STAT_RXFOVERFL))
    {
      outw_p(RP_CHN_STAT_STATMODE, rp->status_reg);
      status |= RP_CHN_STAT_STATMODE;
    }

  if (status & RP_CHN_STAT_STATMODE) 
    {
      int i;

      for (i=0; i<bytes; ++i)
        {
          uint16_t status_rxbyte = inw_p(rp->data_reg);

          rp->cbuf[i] = status_rxbyte & 0xff;

          if (status_rxbyte & RP_CHN_DATA_RXBREAK)
            rp->fbuf[i] = TTY_BREAK;
          else if (status_rxbyte & RP_CHN_DATA_RXFRAME)
            rp->fbuf[i] = TTY_FRAME;
          else if (status_rxbyte & RP_CHN_DATA_RXFOVERFL)
            rp->fbuf[i] = TTY_OVERRUN;
          else if (status_rxbyte & RP_CHN_DATA_RXPARITY)
            rp->fbuf[i] = TTY_PARITY;
          else
            rp->fbuf[i] = TTY_NORMAL;
        }

      // disable status mode?
      if (rx_fifo_count(rp) == 0)
          outw_p(0,rp->status_reg);

      rp_tty_insert_flip_string_flags(tty,rp->cbuf,rp->fbuf,bytes);
    }
  else
    {
      // read data 2 bytes at a time directly into buffer
      insw(rp->data_reg, rp->cbuf, bytes/2);
      if (bytes & 1)
        rp->cbuf[bytes-1] = inb_p(rp->data_reg);
      rp_tty_insert_flip_string(tty,rp->cbuf,bytes);
    }

  rp->port.icount.rx += bytes;
  rp_tty_flip_buffer_push(rp);
}

#if !defined(MIN)
#  define MIN(a,b) ((a)<(b)?(a):(b))
#endif

// function called by polling timer to handle transmit data
static void rp_tx_chars(struct rp_uart_port *rp)
{
  struct circ_buf *xmit = &rp_state(rp)->xmit;
  unsigned total_bytes, contiguous_bytes, txcount, fiforoom;

  if (uart_tx_stopped(&rp->port))
    return;

  fiforoom = TX_FIFO_SIZE - tx_fifo_count(rp);

  while (fiforoom)
    {
      if (rp->port.x_char)
        {
          outb_p(rp->port.x_char, rp->data_reg);
          rp->port.x_char = 0;
          rp->port.icount.tx++;
          --fiforoom;
          continue;
        }

      if (uart_circ_empty(xmit))
        break;

      total_bytes = uart_circ_chars_pending(xmit);
      contiguous_bytes = MIN(total_bytes, (UART_XMIT_SIZE - xmit->tail));
      txcount = MIN(fiforoom, contiguous_bytes);

      // write data 2 bytes at a time directly from xmit buffer
      outsw(rp->data_reg, xmit->buf + xmit->tail, txcount/2);
      if (txcount & 1)
        outb_p(xmit->buf[xmit->tail + txcount - 1], rp->data_reg);

      xmit->tail = (xmit->tail + txcount) & (UART_XMIT_SIZE-1);
      fiforoom -= txcount;
      rp->port.icount.tx += txcount;
    }

  if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
    uart_write_wakeup(&rp->port);
}

static unsigned channel_poll_count[CONFIG_SERIAL_RP_NR_UARTS];

static void rp_channel_poll(struct rp_uart_port *rp)
{
  spin_lock(&rp->port.lock);
  if (!rp->active)
    {
      spin_unlock(&rp->port.lock);
      return;
    }

  ++channel_poll_count[rp->port.line];

  rp_rx_chars(rp);
  rp_tx_chars(rp);

  if (inw_p(rp->intid_reg) & (RP_INTID_DELTA_CD | RP_INTID_DELTA_CTS | RP_INTID_DELTA_DSR))
    wake_up_interruptible(&rp_delta_msr_wait(rp));

  spin_unlock(&rp->port.lock);
}

// We assume rx processor is disabled. However, it has to be enabled
// when we reset the rx fifo count otherwise it will lock up the
// AIOPIC (don't know why).  So, disable the ucode instruction that
// stores data in the rx fifo, enable the rx processor, do the fifo
// flush, then disable the rx processor and restore the ucode
// instruction.
static void rp_flush_rx_fifo(struct rp_uart_port *rp)
{
  inw_p(rp->status_reg);                      //clear pending rx errors

  rp_outb_indexed(rp, RP_RXFIFO_EN, 0x0a);
  set_rx_ctrl_bits(&rp->port, RP_RX_CTRL_RXPROC_EN);

  // reset rx fifo count
  outb_p(RP_CMD_RESRXFCNT | rp->idx, rp->aiop_base+RP_CMD_REG);
  outb_p(                   rp->idx, rp->aiop_base+RP_CMD_REG);

  // clear input/output pointers
  rp_outw_indexed(rp, RP_RXFIFO_OUT_PTR, 0);
  rp_outw_indexed(rp, RP_RXFIFO_IN_PTR, 0);

  clr_rx_ctrl_bits(&rp->port, RP_RX_CTRL_RXPROC_EN);
  rp_outb_indexed(rp, RP_RXFIFO_EN, 0x08);
}

// We assume rx processor is disabled
static void rp_flush_tx_fifo(struct rp_uart_port *rp)
{
  int txEnabled;

  txEnabled = rp->tx_ctrl_value & RP_TX_CTRL_TX_ENABLE;
  if (txEnabled)
    clr_tx_ctrl_bits(&rp->port,  RP_TX_CTRL_TX_ENABLE);

  // reset tx fifo count
  outb_p(rp->idx | RP_CMD_RESTXFCNT, rp->aiop_base+RP_CMD_REG);
  outb_p(rp->idx,                    rp->aiop_base+RP_CMD_REG);

  // clear input/output pointers
  rp_outb_indexed(rp, RP_TXFIFO_OUT_PTR, 0);
  rp_outb_indexed(rp, RP_TXFIFO_IN_PTR, 0);

  if (txEnabled)
    set_tx_ctrl_bits(&rp->port,  RP_TX_CTRL_TX_ENABLE);
}

static int rp_uart_startup(struct uart_port *port)
{
  struct rp_uart_port *rp = port_to_rp(port);
  rp_flush_tx_fifo(rp);
  rp_flush_rx_fifo(rp);
  set_rx_ctrl_bits(&rp->port, RP_RX_CTRL_RXPROC_EN);
  rp->active = 1;
  return 0;
}

static void rp_uart_shutdown(struct uart_port *port)
{
  struct rp_uart_port *rp = port_to_rp(port);
  rp_uart_break_ctl(port, 0);
  rp->active = 0;
}

static const char *rp_uart_type(struct uart_port *port)
{
  if (port->type == PORT_RP)
    {
      struct rp_uart_port *rp = port_to_rp(port);
      static char buffer[1024];
      snprintf(buffer, sizeof buffer, "%s",rp->card->model_info->description);
      return buffer;
    }
  return NULL;
}

static void rp_uart_release_port(struct uart_port *port)
{
  // Nothing to release
}

static int rp_uart_request_port(struct uart_port *port)
{
  // UARTs always present
  return 0;
}

static void rp_uart_config_port(struct uart_port *port, int flags)
{
  if (flags & UART_CONFIG_TYPE)
    port->type = PORT_RP;
}

static int rp_uart_verify_port(struct uart_port *port, struct serial_struct *ser)
{
  if (ser->type != PORT_UNKNOWN && ser->type != PORT_RP)
    return -EINVAL;
  return 0;
}

#define RCKP_RESET_RM2               0x00525005

static int rp_uart_ioctl(struct uart_port *port, unsigned int cmd, unsigned long arg)
{
  switch (cmd) 
    {
    case RCKP_RESET_RM2:
      port_modem_reset(port_to_rp(port));
      break;

    default:
      return -ENOIOCTLCMD;
    }
  return 0;
}


static const struct uart_ops rp_uart_ops =
{
  .tx_empty     = rp_uart_tx_empty,
  .set_mctrl    = rp_uart_set_mctrl,
  .get_mctrl    = rp_uart_get_mctrl,
  .stop_tx      = rp_uart_stop_tx,
  .start_tx     = rp_uart_start_tx,
  .stop_rx      = rp_uart_stop_rx,
  .enable_ms    = rp_uart_enable_ms,
  .break_ctl    = rp_uart_break_ctl,
  .startup      = rp_uart_startup,
  .shutdown     = rp_uart_shutdown,
  .set_termios  = rp_uart_set_termios,
  .type         = rp_uart_type,
  .release_port = rp_uart_release_port,
  .request_port = rp_uart_request_port,
  .config_port  = rp_uart_config_port,
  .verify_port  = rp_uart_verify_port,
  .ioctl        = rp_uart_ioctl,

};

static void rp_reset_card(struct rp_card *card)
{
  int i;
  for (i=0; i<card->model_info->num_aiops; ++i)
    {
      outb_p(RP_CMD_RESET_ALL, card->aiop_base+(i*0x40)+RP_CMD_REG);
      udelay(5);
      outb_p(0,                card->aiop_base+(i*0x40)+RP_CMD_REG);
      udelay(5);
    }
}


static void rp_init_port(struct rp_uart_port *rp, const struct firmware *fw)
{
  int i;

  // reset UART
  outb_p(RP_CMD_RESETUART | rp->idx, rp->aiop_base+RP_CMD_REG);
  outb_p(                   rp->idx, rp->aiop_base+RP_CMD_REG);

  rp_outb_indexed(rp, RP_CLK_PRE, rp->card->model_info->is_plus ?  RP_PLUS_PRESCALE : RP_PRESCALE);
  rp_outw_indexed(rp, RP_TX_CTRL, rp->tx_ctrl_value=0);
  rp_outb_indexed(rp, RP_RX_CTRL, rp->rx_ctrl_value=0);

  // reset rx fifo count and in/out pointers
  outb_p(RP_CMD_RESRXFCNT | rp->idx, rp->aiop_base+RP_CMD_REG);
  outb_p(                   rp->idx, rp->aiop_base+RP_CMD_REG);
  rp_outw_indexed(rp, RP_RXFIFO_OUT_PTR, 0);
  rp_outw_indexed(rp, RP_RXFIFO_IN_PTR, 0);

  // reset tx fifo count and in/out pointers
  outb_p(RP_CMD_RESTXFCNT | rp->idx, rp->aiop_base+RP_CMD_REG);
  outb_p(                   rp->idx, rp->aiop_base+RP_CMD_REG);
  rp_outb_indexed(rp, RP_TXFIFO_OUT_PTR, 0);
  rp_outb_indexed(rp, RP_TXFIFO_IN_PTR, 0);

  // clear tx priority buffer
  rp_outb_indexed(rp, RP_TXPRI_CNT, 0);
  rp_outb_indexed(rp, RP_TXPRI_PTR, 0);
  
  // disable special tx handling
  rp_outb_indexed(rp, RP_TX_ENBLS,  0);

  // receive in data-only mode
  outw_p(0, rp->status_reg);

  // load rx processor microcode
  for (i = 0; i < fw->size; i++)
    rp_outb_indexed(rp, 0x900+i, fw->data[i]);

  __rp_uart_set_termios(rp, CS8 | CREAD | CLOCAL, 0, NULL, 
                        rp->card->model_info->is_plus ? RP_PLUS_DEF_BAUD_DIV : RP_DEF_BAUD_DIV);

  rp_uart_set_mctrl(&rp->port, 0);
}

static void rp_remove_ports(struct rp_card *card)
{
  int i;

  for (i = 0; i < card->initialized_ports; i++)
    uart_remove_one_port(&rp_uart_driver, &card->ports[i].port);
  card->initialized_ports = 0;
}

// called by per-card timer at a frequency set by polling_mode module
// parameter

static void rp_card_poll(unsigned long data)
{
  struct rp_card *card = (void*)data;
  int i;
  if (card->shutting_down)
    return;
  for (i = 0; i < card->n_ports; i++)
    rp_channel_poll(&card->ports[i]);
  mod_timer(&card->poll_timer, jiffies + polling_jiffies);
}

int rp_card_init_ports(const struct firmware *fw, void *context)
{
  struct rp_card *card = context;
  unsigned i;
  int rc = -ENOENT;

  for (i = 0; i < card->n_ports; i++)
    {
      struct rp_uart_port *rp = &card->ports[i];
      struct uart_port *p;
      int ports_per_aiop = card->model_info->is_plus ? 4 : 8; 
      int port_idx = i % ports_per_aiop;
      int aiop_idx = i / ports_per_aiop;
      int line = card->minor_start + i;
      
      rp->card = card;
      rp->aiop_base   = card->aiop_base + (aiop_idx * 0x40);
      rp->aiop_lock   = &card->aiop_lock[aiop_idx];
      rp->data_reg    = rp->aiop_base + RP_CHN_DATA + port_idx*2;
      rp->status_reg  = rp->aiop_base + RP_CHN_STAT + port_idx*2;
      rp->fifocnt_reg = rp->aiop_base + RP_FIFO_CNT + port_idx*2;
      rp->intid_reg   = rp->aiop_base + RP_INT_ID + port_idx;
      rp->index_base  = port_idx * 0x1000;
      rp->idx  = port_idx;

      p = &rp->port;
      p->line = line;
      p->dev = &card->pdev->dev;
      p->type = PORT_RP;
      p->iotype = UPIO_PORT;
      p->uartclk = 16 * (card->model_info->is_plus ? RP_PLUS_BAUD_BASE : RP_BAUD_BASE);
      p->fifosize = 255;          // tx fifo size
      p->ops = &rp_uart_ops;
      p->irq = card->pdev->irq;
      p->iobase = rp->data_reg;

      rp_init_port(rp, fw);

      rc = uart_add_one_port(&rp_uart_driver, p);

      if (rc)
        {
          dev_err(&card->pdev->dev, "error registering port %d: %d\n", i, rc);
          rp_remove_ports(card);
          break;
        }
      card->initialized_ports++;
    }

  dev_info(&card->pdev->dev, "polling at %dHz\n", HZ/polling_jiffies);
  setup_timer(&card->poll_timer, rp_card_poll, (unsigned long)card);
  mod_timer(&card->poll_timer, jiffies+1);
  return 0;
}


static void speaker_reset(struct rp_card *card)
{
  if (card->model_info->modem == 2)
    outb_p(0, card->aiop_base + 0x4f);
  else
    outb_p(0, card->aiop_base + 0x88);
}

static int rp_pci_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
  struct rp_card *card;
  struct rp_uart_port *ports;
  void __iomem * const *bars;
  const struct rp_model_info *model_info;
  int rc;

  //dev_info(&pdev->dev, "found card with device ID %04x\n", id->device);

  model_info = get_model_info(id->device);

  if (!model_info)
    return -EINVAL;

  card = devm_kzalloc(&pdev->dev, sizeof(*card), GFP_KERNEL);
  if (!card)
    return -ENOMEM;
  pci_set_drvdata(pdev, card);

  spin_lock_init(&card->aiop_lock[0]);
  spin_lock_init(&card->aiop_lock[1]);
  spin_lock_init(&card->aiop_lock[2]);
  spin_lock_init(&card->aiop_lock[3]);

  rc = pcim_enable_device(pdev);
  if (rc)
    return rc;

  if (model_info->is_upci)
    rc = pcim_iomap_regions_request_all(pdev, 0x06, DRV_NAME);
  else
    rc = pcim_iomap_regions_request_all(pdev, 0x01, DRV_NAME);
    
  if (rc)
    return rc;

  card->pdev = pdev;
  bars = pcim_iomap_table(pdev);
  //dev_info(&pdev->dev, "bar0=%p bar1=%p bar2=%p pdev=%p\n", bars[0], bars[1], bars[2], card->pdev);

  if (model_info->is_upci)
    {
      card->aiop_base = (unsigned long)(bars[2]);
      card->pci_config_base = (unsigned long)(bars[1]);
    }
  else
    card->aiop_base = (unsigned long)(bars[0]);

  // UPCI quad/octa boards need SW to check which cable is present
  if (model_info->quadocta && model_info->is_upci && !(inw_p(card->pci_config_base+RP_PCI_9030_GPIO_CTRL) & 0x4000))
      model_info = get_model_info(id->device | 0x8000);   // morph into a quad-cable model

  // if it's RMII, check to see if it's a "socketmodem" or not
  if (model_info->modem == 2)
    {
      unsigned class_rev;
      if (!pci_read_config_dword(pdev, PCI_CLASS_REVISION, &class_rev)  && ((class_rev & 0xff) != 1))
        model_info = get_model_info(id->device | 0x4000); // morph into a socketmodem model
    }

  card->model_info = model_info;
  card->n_ports = model_info->num_ports;  // we use this a lot, so elminate a level of indirection

  if (model_info->modem)
    {
      card_modem_reset(card);
      speaker_reset(card);
    }

  card->minor_start = rp_alloc_ports(card->n_ports);

  dev_info(&pdev->dev, "found %s with %d ports\n", card->model_info->description, card->n_ports);

  if (card->minor_start < 0)
    {
      dev_err(&pdev->dev, "too many ports (try increasing CONFIG_SERIAL_RP2_NR_UARTS)\n");
      return -EINVAL;
    }

  rp_reset_card(card);

  ports = devm_kzalloc(&pdev->dev, sizeof(*ports) * card->n_ports, GFP_KERNEL);

  if (!ports)
    return -ENOMEM;

  card->ports = ports;

  return rp_card_init_ports(&rp_firmware,card);
}

static void rp_pci_remove(struct pci_dev *pdev)
{
  struct rp_card *card = pci_get_drvdata(pdev);
  int i = 10;

  card->shutting_down = true;
  msleep(150);  // max poll time is 100ms, after 150ms should be safe
  while (--i)
    {
      if (del_timer_sync(&card->poll_timer)==0)
        break;
      msleep(110);
    }
  if (i==0)
    dev_info(&pdev->dev,"error removing poll timer\n");
  rp_remove_ports(card);
}


static struct pci_driver rp_pci_driver =
{
  .name         = DRV_NAME,
  .id_table     = rp_pci_tbl,
  .probe        = rp_pci_probe,
  .remove       = rp_pci_remove,
};


static int __init rp_module_init(void)
{
  int rc;

  printk(KERN_INFO "rocket: Comtrol RocketPort driver version " DRV_VERS "\n");

  if (HZ < 10)
    {
      printk("rocket: Polling frequency (%dHz) is too low -- using %dHz\n",polling_mode,10);
      polling_mode = 10;
    }

  polling_jiffies = HZ/polling_mode;

  if (polling_jiffies == 0)
    {
      printk("rocket: Polling frequency (%dHz) is too high -- defaulting to %dHz\n",polling_mode,HZ);
      polling_jiffies = 1;
    }

  rc = uart_register_driver(&rp_uart_driver);

  if (rc)
    return rc;

  rc = pci_register_driver(&rp_pci_driver);

  if (rc)
    {
      uart_unregister_driver(&rp_uart_driver);
      printk(KERN_INFO "rocket: error registering PCI driver\n");
      return rc;
    }

  return 0;
}

static void __exit rp_module_exit(void)
{
  printk(KERN_INFO "rocket: unregistering driver\n");
  pci_unregister_driver(&rp_pci_driver);
  uart_unregister_driver(&rp_uart_driver);
}

module_param(polling_mode,int,100);
module_init(rp_module_init);
module_exit(rp_module_exit);

MODULE_DESCRIPTION("Comtrol RocketPort driver");
MODULE_AUTHOR("Comtrol Corp. <support@comtrol.com>");
MODULE_LICENSE("GPL v2");
